/* http-in.c */

/* Copyright 2000 by Eberhard Mattes <em-gw@windhager.de>
   Donated to the public domain.  No warranty.

   Version 0.1	2000-05-27 Initial version
   Version 0.2	2000-05-30 Send only one CRLF after "Connection: close"
   Version 0.3  2000-05-31 Don't terminate until all data has been sent
                2000-06-04 Use common log file format
                2000-08-07 Log longer request lines
                2000-08-12 Don't log copy_bidir() errors twice */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#include <limits.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <fcntl.h>
#include "firewall.h"
#include "libemfw.h"
#include "emio.h"

/* firewall.h lacks some declarations. */

int str_to_port (char *);
int do_daemon (int);
int peername (int, char *, char *, int);
int conn_server (const char *, int, int, char *);

/* Some characters as defined by RFC 1945. */

#define HT      9
#define LF      10
#define CR      13
#define SP      32
#define DEL     127

#define WHITE(c) ((c) == SP || (c) == HT)
#define DIGIT(c) ((c) >= '0' && (c) <= '9')

#define MAX(a,b) ((a) > (b) ? (a) : (b))

static EMO_FILE *c_out;
static EMI_FILE *s_in;
static EMO_FILE *s_out;
static EMO_FILE *log;
static const char *server;
static int client_timeout = 120;
static int server_timeout = 120;
static int overall_timeout;
static const char *log_fname;
static const char *add_header;

/* Configuration. */

static Cfg *confp_global;
static char *proxy_name = "http-in";

/* We reread the configuration after receiving SIGHUP.  This flag is
   set by the signal handler. */

static sig_atomic_t hup_flag;

/* We reopen the log file after receiving SIGUSR1.  This flag is set
   by the signal handler. */

static sig_atomic_t usr1_flag;

/* This flag is true iff -fastdaemon is used. */

static int fastdaemon;
static struct fastdaemon_args fda;

static int major;
static int minor;
static char riaddr[512];
static char buffer[4096];

/* Helper function for error_nolog() and error_log(). */

static void verror (int code, const char *fmt, va_list arg_ptr)
{
  errno = 0;
  if (emo_set_read (c_out) != 0)
    syslog (LLEV, "emo_set_read(): %m");
  emo_printf (c_out, "HTTP/1.0 %d error\r\n", code);
  emo_puts (c_out, "Content-type: text/html\r\n\r\n"
	    "<!DOCTYPE HTML PUBLIC '-//IETF//DTD HTML 2.0//EN'>\n"
	    "<HTML><HEAD><TITLE>Error</TITLE></HEAD>\n"
	    "<BODY><H1>Error</H1>\n");
  emo_vprintf (c_out, fmt, arg_ptr);
  emo_puts (c_out, "\n</BODY></HTML>\n");
  if (emo_flush (c_out) != 0)	/* This sets errno! */
    syslog (LLEV, "error(): %m");
  if (emo_shutdown (c_out) != 0)
    syslog (LLEV, "emo_shutdown(): %m");
  if (emo_read (c_out, 10) != 0)
    syslog (LLEV, "emo_read(): %m");
  exit (1);
}

/* Send an error message to the client (without logging).  CODE is the
   HTTP status code to be returned to the client. */

static void error_nolog (int code, const char *fmt, ...)
{
  va_list arg_ptr;

  va_start (arg_ptr, fmt);
  verror (code, fmt, arg_ptr);
  va_end (arg_ptr);		/* well, verror() doesn't return... */
}

/* Log an error via syslog() and send an error message to the client.
   CODE is the HTTP status code to be returned to the client. */

static void error_log (int code, const char *fmt, ...)
{
  va_list arg_ptr;

  va_start (arg_ptr, fmt);
  vsyslog (LLEV, fmt, arg_ptr);
  va_end (arg_ptr);

  va_start (arg_ptr, fmt);
  verror (code, fmt, arg_ptr);
  va_end (arg_ptr);
}

static void server_write_error (void)
{
    error_log (404, "Cannot send to server: %s", strerror (errno));
}

static void server_write (const void *p, size_t len)
{
  if (emo_write (s_out, p, len) != 0)
    server_write_error ();
}

static void server_writeline (const char *p)
{
  server_write (p, strlen (p));
  server_write ("\r\n", 2);
}

static void server_flush (void)
{
  if (emo_flush (s_out) != 0)
    server_write_error ();
}

/* Read one line terminated by CR*LF (i.e., a LF character preceded by
   any number of CR characters) from F, storing it to the array of
   SIZE characters pointed to by BUF.  The resulting string may
   contain embedded control characters (except for the null character,
   which is rejected).  The terminating CR*LF is removed.  This
   function returns the length of the line. */

static size_t read_line (EMI_FILE *f, char *buf, size_t size, int lineno)
{
  size_t i = 0;
  int c;

  while ((c = emi_getc (f)) != EMI_EOF && c != LF)
    {
      if (i + 1 >= size)
	error_log (414, "header line too long");
      if (c == 0)
	error_log (414, "header line contains invalid character");
      buf[i++] = (char)c;
    }
  if (c == LF)
    {
      /* Ignore any number of CRs. */

      while (i != 0 && buf[i-1] == CR)
        --i;
      buf[i] = 0;
      return i;
    }

  if (emi_error (f))
    syslog (LLEV, "error reading header: %m");
  else
    syslog (LLEV,
	   "premature end of header (line=%d length=%u bytes=%lu)",
	   lineno, (unsigned)i, emi_amount (f, 0));
  exit (1);
}

/* This table defines the supported request methods. */

struct method
{
  const char *name;
  int len;
  unsigned flags;
};

static struct method methods[] =
{
  { "GET",     3},
  { "HEAD",    4},
  { "POST",    4},
  /* Unsupported methods: DELETE, LINK, OPTIONS, PATCH, PUT, TRACE, UNLINK */
};

#define METHODS (sizeof (methods) / sizeof (methods[0]))

int http_method (const char *s, int len)
{
  int m;

  /* Request methods are case-sensitive. */

  for (m = 0; m < METHODS; ++m)
    if (methods[m].len == len
	&& memcmp (s, methods[m].name, len) == 0)
      return 0;
  return -1;
}

/* Parse the HTTP-Version portion of a request header or a
   response-header.  Return the length.  Return -1 on error. */

static int handle_http_version (const char *s)
{
  const char *s0 = s;
  char *end;

  if (strncmp (s, "HTTP/", 5) != 0 || !DIGIT (s[5]))
    return -1;
  s += 5;
  major = strtol (s, &end, 10);
  if (end == s || *end != '.' || !DIGIT (end[1]))
    return -1;
  s = end + 1;
  minor = strtol (s, &end, 10);
  if (end == s || !(*end == 0 || WHITE (*end)))
    return -1;
  if (major != 1)
    return -1;

  /* We support HTTP/1.0 and HTTP/1.1 (TODO) */

  if (minor < 0 || minor > 1)
    return -1;
  return end - s0;
}

/* Parse the request line pointed to by REQ.  This function assumes
   that a null character terminates the line (a null character is
   invalid in a request line anyway). */

static void handle_request_line (const char *req)
{
  int j, url_start, url_end, version_start, version_len;

  /* TODO: Check for control characters */

  /* Process the method. */

  j = 0;
  while (req[j] != 0 && !WHITE (req[j]))
    ++j;
  if (http_method (req, j) != 0)
    error_log (501, "unsupported request method");

  /* Find the URL */

  while (WHITE (req[j]))	/* Be tolerant */
    ++j;
  url_start = j;
  while (req[j] != 0 && !WHITE (req[j]))
    ++j;
  url_end = j;

  /* Find and check the HTTP version. */

  j = strlen (req);
  while (j > 0 && !WHITE (req[j-1]))
    --j;
  version_start = j;
  version_len = handle_http_version (req + j);
  if (version_len < 0 || req[version_start + version_len] != 0)
    error_log (400, "invalid HTTP version");

  /* Don't allow SP in the request URI. */

  j = url_end;
  while (WHITE (req[j]))
    ++j;
  if (j != version_start)
    {
      syslog (LLEV, "Invalid URI in request \"%.512s\"", req);
      error_nolog (400, "invalid URI");
    }
}

static void close_log_file (void)
{
  if (emo_close (log) != 0)
    syslog (LLEV, "Closing log file \"%s\": %m", log_fname);
  log = NULL;
}

static void open_log_file (void)
{
  int fd;

  fd = open (log_fname, O_CREAT | O_APPEND | O_WRONLY, 0644);
  if (fd == -1)
    {
      syslog (LLEV, "Opening log file \"%s\": %m", log_fname);
      exit (1);
    }
  log = emo_fdopen (fd, 16384);
  if (log == NULL)
    {
      syslog (LLEV, "emo_fdopen(): %m");
      exit (1);
    }
}

/* Specific configuration for http-in. */

static void config_http_in (Cfg *confp)
{
  Cfg *cf;
  int i;
  char *new_log_fname = NULL;

  if ((cf = cfg_get ("server", confp)) == (Cfg *)0)
    {
      syslog (LLEV, "fwtkcfgerr: `server' required");
      exit (1);
    }
  config_arg_count (cf, 1);
  server = cf->argv[0];

  if ((cf = cfg_get ("logfile", confp)) != (Cfg *)0)
    {
      config_arg_count (cf, 1);
      new_log_fname = cf->argv[0];
    }

  if (log_fname != NULL && (new_log_fname == NULL
			    || strcmp (log_fname, new_log_fname) != 0))
    close_log_file ();
  log_fname = new_log_fname;
  if (log_fname != NULL && log == NULL)
    open_log_file ();

  add_header = NULL;
  if ((cf = cfg_get ("add-header", confp)) != (Cfg *)0)
    {
      config_arg_count (cf, 1);
      add_header = cf->argv[0];
    }

  i = config_int (confp, "client-timeout", 1, INT_MAX, -1);
  if (i != -1)
      client_timeout = i;

  i = config_int (confp, "server-timeout", 1, INT_MAX, -1);
  if (i != -1)
      server_timeout = i;

  i = config_int (confp, "overall-timeout", 1, INT_MAX, -1);
  if (i != -1)
      overall_timeout = i;
}

static void config_reconfig (void)
{
  if ((confp_global = cfg_read (proxy_name)) == (Cfg *)-1)
    exit (1);
  config_http_in (confp_global);
}

/* Called by do_fastdaemon() on receipt of a signal. */

void got_signal (void)
{
  if (hup_flag)
    {
      hup_flag = 0;
      syslog (LLEV, "reconfiguration: start");
      config_reconfig ();
      syslog (LLEV, "reconfiguration: done");
    }
  if (usr1_flag)
    {
      usr1_flag = 0;
      syslog (LLEV, "reopening log file");
      if (log != NULL)
	close_log_file ();
      if (log_fname != NULL)
	open_log_file ();
    }
}

/* Catch SIGHUP and SIGUSR1.  We just set a flag and let got_signal()
   do the rest. */

void catch_signal (int signo)
{
  if (signo == SIGHUP)
    hup_flag = 1;
  else if (signo == SIGUSR1)
    usr1_flag = 1;
  else if (signo == SIGALRM)
    {
      /* Dirty */
      syslog (LLEV, "timeout");
      exit (1);
    }
}

/* Initializations for -fastdaemon. */

static void init_fastdaemon (void)
{
  struct sigaction sa;

  /* Install (interrupting) signal handler. TODO: Move to library. */

  sa.sa_handler = catch_signal;
  sigemptyset (&sa.sa_mask);
  sa.sa_flags = 0;
#ifdef SA_INTERRUPT
  sa.sa_flags |= SA_INTERRUPT;
#endif
  ALWAYS_ASSERT (sigaction (SIGHUP, &sa, NULL) == 0);
  ALWAYS_ASSERT (sigaction (SIGUSR1, &sa, NULL) == 0);
  ALWAYS_ASSERT (sigaction (SIGALRM, &sa, NULL) == 0);
}

static int is_header (const char *buf, int name_len, const char *s)
{
  return strlen (s) == name_len && strncasecmp (buffer, s, name_len) == 0;
}

/* This is where http-in starts. */

int main (int argc, char *argv[])
{
  EMI_FILE *c_in;
  int fd, r, lines;

  if (argc >= 2 && strcmp (argv[1], "-check") == 0)
    {
      check_snprintf (0);
      exit (0);
    }

  openlog ("http-in", LOG_PID|LOG_NDELAY, LFAC);

  if (argc >= 3 && strcmp (argv[1], "-daemon") == 0)
    {
      int port = str_to_port (argv[2]);
      argc -= 2; argv += 2;
      syslog (LLEV, "Starting daemon mode on port %d", port);
      do_daemon (port);
    }
  else if (argc >= 3 && strcmp (argv[1], "-fastdaemon") == 0)
    {
      parse_fastdaemon (&fda, &argc, &argv);
      fastdaemon = 1;
    }

  if (argc >= 3 && strcmp (argv[1], "-name") == 0)
    {
      proxy_name = argv[2];
      argc -= 2; argv += 2;
    }

  if (argc != 1)
    {
      syslog (LLEV, "too many arguments: %s", argv[1]);
      exit (1);
    }

  config_reconfig ();
  config_groupid (confp_global, 0);
  config_userid (confp_global, 0);

  /* Start daemon mode for -fastdaemon. */

  if (fastdaemon)
    {
      init_fastdaemon ();
      do_fastdaemon (&fda);
    }
  set_userid_groupid ();

  if (overall_timeout > 0)
    alarm (overall_timeout);

  if (fastpeername (0, riaddr, sizeof (riaddr)) != 0)
    {
      syslog (LLEV, "fwtksyserr: cannot get peer name: %m");
      exit (1);
    }

  c_in = emi_fdopen (0, 4096);
  c_out = emo_fdopen (0, 4096);
  emi_timeout (c_in, client_timeout);
  emo_timeout (c_out, client_timeout);

  read_line (c_in, buffer, sizeof (buffer), 0);
  handle_request_line (buffer);

  if (log)
    {
      time_t t;
      char dt[512];

      /* TODO: local time */
      time (&t);
      strftime (dt, sizeof (dt), "%d/%b/%Y:%H:%M:%S +0000", gmtime (&t));

      /* TODO: status code and number of bytes */
      emo_printf (log, "%s \"-\" \"-\" [%s] \"%.8192s\" \"-\" \"-\"\n",
		  riaddr, dt, buffer);
      if (emo_flush (log) != 0)
	{
	  syslog (LLEV, "writing to log file \"%s\": %m", log_fname);
	  exit (1);		/* TODO */
	}
    }

  /* TODO: configure port */
  if ((fd = conn_server (server, 80, 0, (char *)0)) < 0)
    error_log (404, "Cannot connect to server: %s", strerror (errno));

  s_in = emi_fdopen (fd, 4096);
  s_out = emo_fdopen (fd, 4096);
  emi_timeout (s_in, server_timeout);
  emo_timeout (s_out, server_timeout);

  /* Send the request line to the HTTP server. */

  server_writeline (buffer);

  /* Process the request header fields. */

  lines = 0;
  for (;;)
    {
      int name_len, i;
      int drop = 2;		/* 0=keep, 1=drop, 2=drop and log */

      read_line (c_in, buffer, sizeof (buffer), 0);
      if (buffer[0] == 0)
	break;
      if (++lines >= 100)
	{
	  syslog (LLEV, "too many request header lines");
	  exit (1);
	}
      if (!WHITE (buffer[0]))
	{
	  /* Not a continuation line. */

	  i = 0; drop = 2;
	  while (buffer[i] != ':' && !WHITE (buffer[i]))
	    ++i;
	  if (buffer[i] == ':')
	    {
	      name_len = i++;
	      while (WHITE (buffer[i]))
		++i;
	      if (add_header && is_header (buffer, name_len, add_header))
		{
		  /* Drop this header to avoid spoofing. */
		}
	      else if (is_header (buffer, name_len, "Proxy-Connection")
		       || is_header (buffer, name_len, "Connection")
		       || is_header (buffer, name_len, "Keep-Alive"))
		{
		  /* We don't support persistent connections. */
		  drop = 1;
		}
	      else
		drop = 0;
	    }
	}

      /* Note that DROP is initially non-zero, therefore we'll drop
	 leading continuation lines (without logging a message). */

      if (drop == 2)
	syslog (LLEV, "request header line dropped: %.512s", buffer);
      else if (drop == 0)
	server_writeline (buffer);
    }
  if (add_header)
    emo_printf (s_out, "%s: %s\r\n", add_header, riaddr);
  if (major == 1 && minor == 1)
    server_writeline ("Connection: close");
  server_write ("\r\n", 2);
  server_flush ();

  /* TODO: Handle request body */
  /* TODO: Handle response */
  /* TODO: Send "Connection: close" for HTTP/1.1 */
  /* TODO: Persistent connections */
  /* TODO: SO_LINGER? */

  /* CB_EOF2 here means: terminate as soon as the server shuts down
     sends (closes the socket), don't terminate when the client shuts
     down sends. */

  r = copy_bidir_emi (c_in, -1, NULL, fd,
		      MAX (client_timeout, server_timeout), CB_EOF2);
  if (r > 0)
    syslog (LLEV, "timeout");
  return 0;
}
