/* netpermd.c */

/* Copyright 2000 by Eberhard Mattes <em-gw@windhager.de>
   Donated to the public domain.  No warranty.

   Version 0.1	2000-04-09 Initial version
   Version 0.2	2000-12-11 Add classes */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include "firewall.h"
#include "libemfw.h"
#include "libemtn.h"

int enargv (char *buf, char **av, int avsiz, char *tobuf, int bsiz);

#define CFG_FNAME       "/usr/local/etc/netperm-table"
#define TMP_FNAME       "/usr/local/etc/netperm-table.tmp"
#define LOCK_FNAME      "/usr/local/etc/netperm-table.lock"
#define USER_FNAME      "/usr/local/etc/netpermd.cfg"
#define PID_FNAME	"/var/run/squid-gw.pid"

#define ISBLANK(c) ((c) == ' ' || (c) == '\t')

struct user_actions
{
  int connect;
  int destinations;
  char classes[256];
};

static struct user_actions cur_ua;

static int lock_cfg (tnconn *t)
{
  int fd, e;

  for (;;)
    {
      fd = open (LOCK_FNAME, O_CREAT | O_WRONLY | O_EXCL);
      if (fd >= 0)
        break;
      e = errno;
      syslog (LOG_NOTICE, "%s: %m", LOCK_FNAME);
      if (e != EEXIST)
        {
          tn_printf (t, "%s: %s\r\n", LOCK_FNAME, strerror (e));
          return -1;
        }
      tn_puts (t, "netperm-table locked, please wait...\r\n");
      sleep (1);
    }
  close (fd);
  return 0;
}

static void unlock_cfg (void)
{
  unlink (LOCK_FNAME);
}

static int check_user (const char *fname, const char *user, const char *passwd)
{
  FILE *f;
  char line[512];
  char obuf[512];
  char *av[16];
  char *e;
  int lineno = 0;
  int rc = 1;
  int ac, i;
  size_t len, ci;
  struct user_actions tmp_ua;

  f = fopen (fname, "r");
  if (f == NULL)
    {
      syslog (LOG_NOTICE, "%s: %m", fname);
      return -1;
    }
  while (fgets (line, sizeof (line), f) != NULL)
    {
      ++lineno;
      e = strchr (line, '\n');
      if (e == NULL)
        {
          syslog (LOG_NOTICE, "%s:%d: line too long or not terminated",
                  fname, lineno);
          fclose (f);
          return -1;
        }
      *e = 0;
      if (line[0] == 0 || line[0] == '#')
        continue;

      ac = enargv (line, av, sizeof (av) / sizeof (av[0]),
		   obuf, sizeof (obuf));
      if (ac < 3)		/* includes ac < 0 */
	goto syntax;
      i = 2;
      tmp_ua.connect = 0;
      tmp_ua.destinations = 0;
      tmp_ua.classes[0] = 0;
      ci = 0;
      while (i < ac)
	{
	  if (strcmp (av[i], "-connect") == 0)
	    {
	      tmp_ua.connect = 1;
	      ++i;
	    }
	  else if (strcmp (av[i], "-destinations") == 0)
	    {
	      tmp_ua.destinations = 1;
	      ++i;
	      while (i < ac && av[i][0] != '-')
		{
		  if (strpbrk (av[i], " \t") != NULL)
		    goto syntax;
		  len = strlen (av[i]);
		  if (1 + ci + len >= sizeof (tmp_ua.classes))
		    goto syntax;
		  tmp_ua.classes[ci++] = ' ';
		  memcpy (tmp_ua.classes + ci, av[i], len + 1);
		  ci += len;
		  ++i;
		}
	    }
	  else
	    goto syntax;
	}
      if (!tmp_ua.connect && !tmp_ua.destinations)
	goto syntax;
      if (strcmp (user, av[0]) == 0 && strcmp (passwd, av[1]) == 0)
	{
	  rc = 0;
	  cur_ua = tmp_ua;
	}
    }
  if (ferror (f))
    {
      syslog (LOG_NOTICE, "%s: %m", fname);
      fclose (f);
      return -1;
    }
  fclose (f);
  return rc;

syntax:
  syslog (LOG_NOTICE, "%s:%d: syntax error", fname, lineno);
  fclose (f);
  return -1;
}

/* TODO: IP address */
static int check_hostname (const char *s)
{
  for (;;)
    {
      const char *s0 = s;
      while (*s != 0
             && (strchr ("ABCDEFGHIJKLMNOPQRSTUVWXYZ"
                         "abcdefghijklmnopqrstuvwxyz", *s) != NULL
                 || (s != s0 && strchr ("0123456789-", *s) != NULL)))
        ++s;
      if (s - s0 == 0 || s - s0 > 63 || s[-1] == '-')
        return -1;
      if (*s == 0)
        return 0;
      if (*s != '.')
        return -1;
      ++s;
    }
}

static const char *config_error (int lineno, const char *msg)
{
  static char buf[1024];
  snprintf (buf, sizeof (buf), "%s:%d: %s", "netperm-table", lineno, msg);
  return buf;
}

static const char *extract_classes (char *out, size_t out_size,
				    char **av, int ac)
{
  int i = 2;
  char *d = out;
  size_t len;

  while (i < ac)
    if (strcmp (av[i++], "-class") == 0)
      {
	if (i >= ac || av[i][0] == '-')
	  return "missing class name";
	len = strlen (av[i]);
	if (d - out + 1 + len >= out_size)
	  return "too many classes";
	*d++ = ' ';
	memcpy (d, av[i], len);
	d += len;
	*d = 0;
	++i;
      }
  return NULL;
}

enum do_config_mode
{
  dcm_connect,
  dcm_destinations
};

static const char *do_config2 (const char *line, int lineno,
			       enum do_config_mode mode, const char *value,
			       char *out, size_t out_size, FILE *df)
{
  char obuf[BUFSIZ];
  char *av[256];
  int ac, deny;
  const char *s, *msg;

  ac = enargv (strchr (line, ':') + 1,
	       av, sizeof (av) / sizeof (av[0]),
	       obuf, sizeof (obuf));
  if (ac < 0)
    return config_error (lineno, "too many arguments");
  if (ac < 1)
    return config_error (lineno, "too few arguments");
  s = av[0];
  deny = 0;
  if (strncmp (s, "permit-", 7) == 0)
    s += 7;
  else if (strncmp (s, "deny-", 5) == 0)
    {
      s += 5;
      deny = 1;
    }
  if (mode == dcm_connect && strcmp (s, "connect") == 0)
    {
      if (ac < 2)
	return config_error (lineno, "too few arguments");
      if (strcasecmp (av[1], value) == 0)
	return "Rule already exists";
    }
  if (mode == dcm_destinations && strcmp (s, "destinations") == 0)
    {
      if (ac < 2)
	return config_error (lineno, "too few arguments");
      /* TODO: url_compare()? */
      if (strcasecmp (av[1], value) == 0)
	{
	  if (out != NULL)
	    {
	      ALWAYS_ASSERT (out_size >= 2);
	      if (deny)
		strcpy (out, "-"); /* deny-destinations */
	      else
		{
		  strcpy (out, " "); /* permit-destinations */
		  msg = extract_classes (out, out_size, av, ac);
		  if (msg != NULL)
		    return msg;
		}
	    }
	  if (df != NULL)
	    fputc ('#', df);	/* Disable the original line */
	}
    }
  return NULL;			/* Success */
}

static const char *do_config1 (enum do_config_mode mode,
			       const char *newcfg, const char *value,
			       char *out, size_t out_size,
			       FILE *sf, FILE *df)
{
  char line[BUFSIZ];
  char *e;
  const char *s, *msg;
  int found, lineno;

  lineno = 0;
  found = 0;
  while (fgets (line, sizeof (line), sf) != NULL)
    {
      ++lineno;
      e = strchr (line, '\n');
      if (e == NULL)
        return config_error (lineno, "line too long or not terminated");
      *e = 0;
      if (newcfg != NULL && !found
	  && strcmp (line, "# netpermd: squid-gw end") == 0)
        {
          fprintf (df, "%s\n", newcfg);
          found = 1;
        }
      s = line;
      while (ISBLANK (*s))
        ++s;
      if (*s != 0 && *s != '#')
	{
	  s = strchr (line, ':');
	  if (s == NULL)
	    return config_error (lineno, "missing colon");

	  while (s != line && ISBLANK (s[-1]))
	    --s;
	  if (s - line == 8 && strncmp (line, "squid-gw", 8) == 0)
	    {
	      msg = do_config2 (line, lineno, mode, value, out, out_size, df);
	      if (msg != NULL)
		return msg;
	    }
        }
      fprintf (df, "%s\n", line);
    }
  if (ferror (sf))
    return "netperm-table: read error";

  if (df != NULL && !found)
    return "netperm-table: missing section mark";
  return NULL;			/* Success */
}

static const char *do_config (enum do_config_mode mode,
			      const char *newcfg, const char *value,
			      char *out, size_t out_size)
{
  FILE *sf, *df;
  const char *msg;

  if (out != NULL)
    out[0] = 0;
  sf = fopen (CFG_FNAME, "r");
  if (sf == NULL)
    return "Cannot open netperm-table";

  msg = do_config1 (mode, newcfg, value, out, out_size, sf, NULL);
  if (msg != NULL)
    {
      fclose (sf);
      return msg;
    }
  if (newcfg == NULL)
    {
      if (fclose (sf) != 0)
	return "netperm-table: read error";
      return NULL;		/* Success */
    }

  df = fopen (TMP_FNAME, "w");
  if (df == NULL)
    {
      fclose (sf);
      return "Cannot open temporary file";
    }
  rewind (sf);
  msg = do_config1 (mode, newcfg, value, out, out_size, sf, df);
  if (msg != NULL)
    {
      fclose (sf);
      remove (TMP_FNAME);
      fclose (df);
      return msg;
    }

  if (fclose (sf) != 0)
    {
      remove (TMP_FNAME);
      fclose (df);
      return "netperm-table: read error";
    }

  if (fclose (df) != 0)
    return "Cannot write temporary file";

  /* Unfortunately, a time window is unavoidable here. */
  if (rename (TMP_FNAME, CFG_FNAME) != 0)
    return "Cannot rename temporary file";

  return NULL;			/* Success */
}

static void kill_squid_gw (tnconn *t)
{
  FILE *f;
  char line[40];
  char *e;
  long pid;

  f = fopen (PID_FNAME, "r");
  if (f == NULL)
    {
      syslog (LOG_NOTICE, "%s: %m", PID_FNAME);
      tn_puts (t, "Cannot open PID file of squid-gw.\r\n");
      return;
    }

  if (fgets (line, sizeof (line), f) == NULL)
    {
      syslog (LOG_NOTICE, "%s: read failure", PID_FNAME);
      tn_puts (t, "Cannot read PID file of squid-gw.\r\n");
    }
  else
    {
      pid = strtol (line, &e, 10);
      if (pid < 1 || *e != '\n' || (pid_t)pid != pid)
	{
	  syslog (LOG_NOTICE, "%s: no PID", PID_FNAME);
	  tn_puts (t, "PID file of squid-gw does not contain a PID.\r\n");
	}
      else if (kill ((pid_t)pid, SIGHUP) != 0)
	{
	  syslog (LOG_NOTICE, "kill(%ld): %m", (long)pid);
	  tn_puts (t, "Sending SIGHUP to squid-gw failed.\r\n");
	}
      else
	{
	  syslog (LOG_NOTICE, "SIGHUP sent to %ld", (long)pid);
	  tn_puts (t, "SIGHUP sent to squid-gw.\r\n");
	}
    }
  fclose (f);
}

static int change_config (tnconn *t, enum do_config_mode mode,
			  const char *newcfg, const char *value)
{
  const char *msg;

  if (lock_cfg (t) != 0)
    return -1;
  msg = do_config (mode, newcfg, value, NULL, 0);
  unlock_cfg ();
  if (msg != NULL)
    {
      syslog (LOG_NOTICE, "%s", msg);
      tn_printf (t, "%s.\r\n", msg);
      return -1;
    }
  else
    {
      syslog (LOG_NOTICE, "added: %s", newcfg);
      tn_printf (t, "\r\nDone.  The rule line\r\n\r\n"
		 "  %s\r\n\r\n"
		 "has been added to netperm-table.\r\n\r\n", newcfg);
      kill_squid_gw (t);
      return 0;
    }
}

static int config_connect1 (tnconn *t)
{
  char hostname[255+1];
  char buf[4096];
  struct hostent *he;
  char **p;
  char *d;

  tn_puts (t, "\r\nHostname: ");
  tn_gets (t, hostname, sizeof (hostname), 120, TN_GETS_ECHO);
  if (strcmp (hostname, "q") == 0)
    return 1;
  if (check_hostname (hostname) != 0)
    {
      syslog (LOG_NOTICE, "invalid hostname: %s", hostname);
      tn_puts (t, "Invalid hostname.\r\n");
      return -1;
    }

  he = gethostbyname (hostname);
  if (he == NULL)
    {
      syslog (LOG_NOTICE, "gethostbyname: h_errno=%d", h_errno);
      tn_printf (t, "Getting IP address failed, h_errno=%d.\r\n",
                 h_errno);
      return -1;
    }
  if (he->h_addrtype != AF_INET)
    {
      syslog (LOG_NOTICE, "h_addrtype is %d", (int)he->h_addrtype);
      tn_puts (t, "h_addrtype is not AF_INET.\r\n");
      return -1;
    }
  if (he->h_length != 4)
    {
      syslog (LOG_NOTICE, "h_length is %d", (int)he->h_length);
      tn_printf (t, "h_length is not 4.\r\n");
      return -1;
    }
  p = he->h_addr_list;
  if (p == NULL)
    {
      syslog (LOG_NOTICE, "no IP address");
      tn_printf (t, "No IP address.\r\n");
      return -1;
    }

  /* TODO: Port */

  d = buf;
  d += sprintf (d, "squid-gw: connect %s -ipaddr", hostname);
  while (*p != NULL)
    {
      const char *ipaddr = inet_ntoa (*(struct in_addr *)*p++);
      if (d - buf + 1 + strlen (ipaddr) >= sizeof (buf))
        {
	  syslog (LOG_NOTICE, "too many IP addresses");
          tn_puts (t, "Too many IP addresses.\r\n");
          return -1;
        }
      d += sprintf (d, " %s", ipaddr);
    }

  return change_config (t, dcm_connect, buf, hostname);
}

static void config_connect (tnconn *t)
{
  tn_puts (t, "\r\n"
           "Please enter the hostname for which\r\n"
           "  https://HOSTNAME/...\r\n"
           "should be permitted.  Type q to leave.\r\n");
  while (config_connect1 (t) <= 0)
    ;
}

static int find_class (const char *s, size_t n, const char *classes,
		       int with_option)
{
  const char *e;
  while (*classes != 0)
    {
      ALWAYS_ASSERT (*classes == ' ');
      ++classes;
      if (with_option)
	{
	  if (strncmp (classes, "-class ", 7) != 0)
	    abort ();
	  classes += 7;
	}
      e = classes;
      while (*e != 0 && !ISBLANK (*e))
	++e;
      if (e - classes == n && memcmp (s, classes, n) == 0)
	return 0;
      classes = e;
    }
  return -1;
}

static int check_classes (tnconn *t, const char *s)
{
  const char *e;
  int len;

  for (;;)
    {
      while (ISBLANK (*s))
	++s;
      if (*s == 0)
	break;
      e = s;
      while (*e != 0 && !ISBLANK (*e))
	++e;
      len = e - s;
      if (find_class (s, len, cur_ua.classes, 0) != 0)
	{
	  tn_printf (t, "Configuration class \"%.*s\" is not available.\r\n",
		     len, s);
	  return -1;
	}
      s = e;
    }
  return 0;
}

static int add_classes (char *dst, size_t dst_size,
			const char *classes, const char *s)
{
  const char *e;
  int len;
  char *d = dst + strlen (dst);

  for (;;)
    {
      while (ISBLANK (*s))
	++s;
      if (*s == 0)
	break;
      e = s;
      while (*e != 0 && !ISBLANK (*e))
	++e;
      len = e - s;
      if (find_class (s, len, classes, 1) != 0)
	{
	  if (d - dst + 8 + len >= dst_size)
	    return -1;
	  memcpy (d, " -class ", 8);
	  d += 8;
	  memcpy (d, s, len);
	  d += len;
	  *d = 0;		/* Do this now, for find_class() above! */
	}
      s = e;
    }
  return 0;
}

static int check_url (const char *url)
{
  if (strncmp (url, "http://", 7) != 0)
    return -1;
  url += 7;
  if (*url == 0)
    return -1;
  /* TODO: move squid-gw/url.c to libem */
  while (*url != 0)
    {
      unsigned char c = *url++;
      if (c < 0x20 || strchr ("* \t#\"'", c) != NULL)
	return -1;
    }
  return 0;
}

static int empty (const char *s)
{
  while (*s != 0)
    {
      if (!ISBLANK (*s))
	return 0;
      ++s;
    }
  return 1;
}

static int config_destinations1 (tnconn *t)
{
  char url[255+1];
  char buf[4096];
  char old_classes[256];
  char new_classes[256];
  const char *msg;
  char *d;

  tn_puts (t, "\r\nURL: ");
  tn_gets (t, url, sizeof (url), 120, TN_GETS_ECHO);
  if (strcmp (url, "q") == 0)
    return 1;
  if (check_url (url) != 0)
    {
      syslog (LOG_NOTICE, "invalid URL: %s", url);
      tn_puts (t, "Invalid URL.\r\n");
      return -1;
    }

  if (lock_cfg (t) != 0)
    return -1;
  msg = do_config (dcm_destinations, NULL, url,
		   old_classes, sizeof (old_classes));
  unlock_cfg ();
  if (msg != NULL)
    {
      syslog (LOG_NOTICE, "%s", msg);
      tn_printf (t, "%s.\r\n", msg);
      return -1;
    }

  /* Possible values of OLD_CLASSES:

     ""		No matching line found
     "-"	Matching deny-destinations line found
     " "	Matching permit-destinations line found, without -class
     " X Y"	Matching permit-destinations line found, -class X -class Y */

     
  if (old_classes[0] == '-')
    {
      tn_printf (t, "There is a \"deny-destinations\" line for "
		 "that URL.\r\n");
      return -1;
    }
  if (old_classes[0] == 0)
    tn_printf (t, "There is not yet a \"destinations\" line for "
	       "that URL.\r\n");
  else if (old_classes[1] == 0)
    tn_printf (t, "There is already a \"destinations\" line for "
	       "that URL,\r\n without any configuration classes\r\n");
  else
    tn_printf (t, "There is already a \"destinations\" line for "
	       "that URL,\r\n with these configuration classes: %s\r\n",
	       old_classes + 1);

  new_classes[0] = 0;
  if (cur_ua.classes[0] != 0)
    {
      tn_printf (t, "\r\nPlease enter the configuration classes (if any) "
		 "to be added,\r\n"
		 "separated by blanks. Type q to leave.\r\n"
		 "These classes are available: %s\r\n", cur_ua.classes + 1);
      for (;;)
	{
	  tn_puts (t, "\nAdd classes: ");
	  tn_gets (t, new_classes, sizeof (new_classes), 120, TN_GETS_ECHO);
	  if (strcmp (new_classes, "q") == 0)
	    return -1;		/* Ask for another URL */
	  if (check_classes (t, new_classes) == 0)
	    break;
	}
    }

  /* TODO: copy other options from original line */

  d = buf;
  d += sprintf (d, "squid-gw: permit-destinations %s", url);
  if (!empty (old_classes) || !empty (new_classes))
    {
      if (add_classes (buf, sizeof (buf), d, old_classes) != 0 ||
	  add_classes (buf, sizeof (buf), d, new_classes) != 0)
	{
	  tn_puts (t, "\r\nToo many classes.\r\n");
	  return -1;
	}
    }
  return change_config (t, dcm_destinations, buf, url);
}

static void config_destinations (tnconn *t)
{
  tn_puts (t, "\r\n"
           "Please enter the URL for which you want to add configuration\r\n"
	   "classes.  Type q to leave.\r\n");
  while (config_destinations1 (t) <= 0)
    ;
}

int main (int argc, char *argv[])
{
  char user[32];
  char passwd[32];
  char cmd[2];
  tnconn *t;

  if (argc >= 2 && strcmp (argv[1], "-check") == 0)
    {
      check_snprintf (0);
      exit (0);
    }

  openlog ("netpermd", LOG_PID|LOG_NDELAY, LOG_DAEMON);

  t = tn_init (0, TN_INIT_TELNET);
  if (t == NULL)
    {
      syslog (LOG_NOTICE, "tn_init() failed");
      exit (2);
    }
  tn_puts (t, "\r\n"
	   "netperm-table configuration\r\n"
	   "===========================\r\n");
  for (;;)
    {
      tn_puts (t, "\r\nLogin: ");
      tn_gets (t, user, sizeof (user), 120, TN_GETS_ECHO);
      tn_puts (t, "Password: ");
      tn_gets (t, passwd, sizeof (passwd), 120, 0);
      sleep (1);
      if (check_user (USER_FNAME, user, passwd) == 0)
        break;
      syslog (LOG_NOTICE, "access denied: %s/%s", user, passwd);
      tn_puts (t, "Access denied.\r\n");
      sleep (4);
    }
  syslog (LOG_NOTICE, "access permitted: %s", user);

  for (;;)
    {
      tn_puts (t, "\r\nThe following commands are available:\r\n\r\n");
      if (cur_ua.connect)
	tn_puts (t, "  c  Add a \"connect\" line (for https)\r\n");
      if (cur_ua.destinations)
	tn_puts (t, "  d  Add a \"destinations\" line\r\n");
      tn_puts (t, "  q  Quit\r\n");
      tn_puts (t, "\r\nCommand: ");
      tn_gets (t, cmd, sizeof (cmd), 120, TN_GETS_ECHO);
      if (strcmp (cmd, "c") == 0 && cur_ua.connect)
	config_connect (t);
      else if (strcmp (cmd, "d") == 0 && cur_ua.destinations)
	config_destinations (t);
      else if (strcmp (cmd, "q") == 0)
	break;
      else
	tn_puts (t, "\r\nInvalid command.\r\n");
    }

  syslog (LOG_NOTICE, "terminated");
  tn_puts (t, "Good bye.\r\n");
  exit (0);
}
