/* nntp-gw.c */

/* Copyright 1997-2000 by Eberhard Mattes <em-gw@windhager.de>
   Copyright 1999-2000 by Giles Heron <giles@clear.net.nz>
   Donated to the public domain.  No warranty.

   Version 0.1	1997-09-03 Initial version
   Version 0.2	1998-09-03 Support XHDR and XPAT; check_snprintf()
   Version 0.3	1999-07-09 LIST EXTENSIONS, etc.; <KEY>; server port
   Version 0.4	2000-04-17 Add G. Heron's multiple newsserver & auth support
   Version 0.5	2000-07-17 Fix various bugs */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <signal.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/file.h>
#include "firewall.h"
#include "libemfw.h"
#include "emio.h"

/* Reference: RFC 977 */

#define SERVER_PORT	119  /* Default port for NNTP */

#define LF	10
#define CR	13

#define WHITE(c) ((c) == 32 || (c) == 9)

/* Flags in command table. */

#define CMD_SIZE	1	/* Log list size (with "log list-size") */
#define CMD_NOAUTH	2	/* No authorization required */

static Cfg *confp;
static const char *user;
static const char *server;
static int server_port = SERVER_PORT;
static char riaddr[512];
static char rladdr[512];
static char input[512];
static char cmd[32];
static size_t input_length;
static size_t cmd_length;
static int server_fd;
static int server_contacted;
static EMI_FILE *c_in, *s_in;
static EMO_FILE *c_out, *s_out;
static char group_name[512];
static unsigned long group_bytes;
static unsigned long group_articles;
static int log_auth;
static int log_command;
static int log_response;
static int log_list_size;
static int log_post;
static int log_article_summary;

/* firewall.h lacks some declarations. */

int conn_server (const char *, int, int, char *);
int str_to_port (char *);
int do_daemon (int);
int peername (int, char *, char *, int);
int hostmatch (const char *, const char *);

/* Prototypes. */

static void quit (int) ATTR_NORETURN;
static void read_error (EMI_FILE *) ATTR_NORETURN;
static void write_error (EMO_FILE *) ATTR_NORETURN;

static void log_group (void)
{
  if (log_article_summary && group_name[0] != 0 && group_articles != 0)
    syslog (LLEV, "group %.256s: articles=%lu bytes=%lu",
            group_name, group_articles, group_bytes);
}

static void quit (int rc)
{
  log_group ();
  syslog (LLEV, "exit host=%.512s/%.20s "
          "client_in=%lu client_out=%lu server_in=%lu server_out=%lu",
          rladdr, riaddr,
          emi_amount (c_in, 0), emo_amount (c_out, 0),
          emi_amount (s_in, 0), emo_amount (s_out, 0));
  _exit (rc);
}

/* Return "client" or "server", depending on whether the file F is
   used for the client or server, respectively. */

static const char *cs (void *f)
{
  return (f == c_in || f == c_out) ? "client" : "server";
}

static void read_error (EMI_FILE *f)
{
  syslog (LLEV, "%s: read(): %m", cs (f));
  quit (1);
}

static void write_error (EMO_FILE *f)
{
  syslog (LLEV, "%s: write(): %m", cs (f));
  quit (1);
}

static size_t read_line (EMI_FILE *f, char *buf, size_t size)
{
  size_t i = 0;
  int c;

  while ((c = emi_getc (f)) != EMI_EOF && c != LF)
    {
      if (i + 1 >= size)
        {
          syslog (LLEV, "%s: line too long", cs (f));
          quit (1);
        }
      if (c == 0)
        {
          syslog (LLEV, "%s: line contains invalid character", cs (f));
          quit (1);
        }
      buf[i++] = (char)c;
    }
  if (c == LF)
    {
      if (i != 0 && buf[i-1] == CR)
        --i;
      buf[i] = 0;
      return i;
    }
  if (emi_error (f))
    read_error (f);
  else
    {
      syslog (LLEV, "%s: unterminated line", cs (f));
      quit (1);
    }
}

static void command (void)
{
  if (log_command)
    syslog (LLEV, "command: %.512s", input);
  emo_write (s_out, input, input_length);
  emo_write (s_out, "\r\n", 2);
  if (emo_flush (s_out) != 0)
    write_error (s_out);
}

static void respond (const char *s)
{
  emo_puts (c_out, s);
  emo_write (c_out, "\r\n", 2);
  if (emo_flush (c_out) != 0)
    write_error (c_out);
}

static int has_body (int code)
{
  return (code == 100 || code == 215 || code == 220 || code == 221
          || code == 222 || code == 224 || code == 230 || code == 231);
}

static int get_response (void)
{
  int code, i;

  input_length = read_line (s_in, input, sizeof (input));
  if (log_response)
    syslog (LLEV, "response: %.512s", input);
  code = 0;
  for (i = 0; i < 3; ++i)
    {
      if (input[i] >= '0' && input[i] <= '9')
        code = code * 10 + input[i] - '0';
      else
        {
          code = -1;
          break;
        }
    }
  return code;
}

static void put_response (void)
{
  emo_write (c_out, input, input_length);
  emo_write (c_out, "\r\n", 2);
  if (emo_flush (c_out) != 0)
    write_error (c_out);
}

static int copy_response (void)
{
  int code = get_response ();
  put_response ();
  return code;
}

static unsigned long copy_body (EMO_FILE *dst, EMI_FILE *src)
{
  int c;
  enum
  {
    STATE_CR,                   /* CR */
    STATE_CR_LF,                /* CR LF */
    STATE_DOT,                  /* CR LF . */
    STATE_DOT_CR,               /* CR LF . CR */
    STATE_DOT_CR_LF,            /* CR LF . CR LF */
    STATE_OTHER                 /* everything else */
  } state;
  unsigned long n0;

  n0 = emi_amount (src, 0);
  state = STATE_CR_LF;
  for (;;)
    {
      /* We're in state STATE_CR or STATE_CR_LF now. */

      do
        {
          c = emi_getc (src);
          if (c == EMI_EOF)
            break;
          emo_putc (dst, c);
          switch (state)
            {
            case STATE_CR:
              if (c == LF)
                state = STATE_CR_LF;
              else if (c == CR)
                state = STATE_CR;
              else
                state = STATE_OTHER;
              break;
            case STATE_CR_LF:
              if (c == '.')
                state = STATE_DOT;
              else if (c == LF)
                state = STATE_CR_LF;
              else if (c == CR)
                state = STATE_CR;
              else
                state = STATE_OTHER;
              break;
            case STATE_DOT:
              if (c == LF)
                state = STATE_DOT_CR_LF;
              else if (c == CR)
                state = STATE_DOT_CR;
              else
                state = STATE_OTHER;
              break;
            case STATE_DOT_CR:
              if (c == LF)
                state = STATE_DOT_CR_LF;
              else if (c == CR)
                state = STATE_CR;
              else
                state = STATE_OTHER;
              break;
            default:
              syslog (LLEV, "internal error");
              quit (1);
            }
        } while (state != STATE_DOT_CR_LF && state != STATE_OTHER);

      if (c == EMI_EOF || state == STATE_DOT_CR_LF)
        break;

      do
        {
          c = emi_getc (src);
          if (c == EMI_EOF)
            break;
          emo_putc (dst, c);
        } while (c != LF && c != CR);
      if (c == LF)
        state = STATE_CR_LF;
      else
        state = STATE_CR;
    }

  if (c == EMI_EOF)
    {
      syslog (LLEV, "%s: unterminated line", cs (src));
      quit (1);
    }
  if (emo_flush (dst) != 0)
    write_error (dst);
  return emi_amount (src, 0) - n0;
}

static int check_group (const char *group, const char **msg, int *delay)
{
  Cfg *cf = cfg_get ("newsgroups", confp);
  int i;
  size_t len;

  len = strlen (group);
  while (cf != NULL)
    {
      for (i = 0; i < cf->argc && cf->argv[i][0] != '-'; ++i)
	{
	  if (lower_match (cf->argv[i], strlen (cf->argv[i]), group, len))
            {
	      if (cf->flags & PERM_DENY)
		goto deny;
	      return 0;
	    }
	}
      cf = cfg_get ("newsgroups", (Cfg *)0);
    }
  /* permit by default */
  return 0;

deny:
  syslog (LLEV, "denied group=%.512s", group);
  *msg = "411 access denied";
  *delay = 0;
  while (++i < cf->argc)
    {
      if (i + 1 < cf->argc && strcmp (cf->argv[i], "-message") == 0)
	*msg = cf->argv[++i];
      else if (i + 1 < cf->argc && strcmp (cf->argv[i], "-delay") == 0)
	*delay = atoi (cf->argv[++i]);
      else if (strcmp (cf->argv[i], "-quit") == 0)
	*delay = -1;
    }
  return -1;
}

static int config_destinations (const char *server_name, Cfg *confp)
{
  Cfg *cf = cfg_get ("destinations", confp);
  int i;

  while (cf != NULL)
    {
      for (i = 0; i < cf->argc && cf->argv[i][0] != '-'; ++i)
        {
          if (hostmatch (cf->argv[i], server_name))
            {
              if (cf->flags & PERM_DENY)
                goto deny;
	      goto match;
            }
        }
      cf = cfg_get ("destinations", (Cfg *)0);
    }
deny:
  syslog (LLEV, "deny use of news server %s", server_name);
  return 0;

match:
  while (i < cf->argc && cf->argv[i][0] != '-')
    ++i;
  while (i < cf->argc)
    {
      if (i + 1 < cf->argc && strcmp (cf->argv[i], "-port") == 0)
	server_port = str_to_port (cf->argv[++i]);
      else
	{
	  syslog (LLEV, "fwtkcfgerr: invalid option in line %d", cf->ln);
	  exit (1);
	}
      ++i;
    }

  syslog (LLEV, "permit use of news server %s port %d",
	  server_name, server_port);
  return 1;
}

/* Get banner from remote server and pass back. */

static void display_banner (void)
{
  int code;

  code = copy_response ();
  if (code != 200 && code != 201)
    quit (1);
}

/* Get banner from remote server and discard. */

static void discard_banner (void)
{
  char banner[512];

  read_line (s_in, banner, sizeof (banner));
  if (log_response)
    syslog (LLEV, "discarded banner: %.512s", input);
}

/* Open a connection to the server. */

static void open_server (void)
{
  if ((server_fd = conn_server (server, server_port, 0, (char *)0)) < 0)
    {
      syslog (LLEV, "cannot connect to server %.512s/%d: %m",
	      server, server_port);
      respond ("400 cannot connect to server");
      exit (1);
    }
  s_in = emi_fdopen (server_fd, 4096);
  s_out = emo_fdopen (server_fd, 4096);
  if (s_in == NULL || s_out == NULL)
    {
      syslog (LLEV, "fdopen() failed: %m");
      exit (1);
    }
  emi_timeout (s_in, PROXY_TIMEOUT);
  emo_timeout (s_out, PROXY_TIMEOUT);

  /* Scary side-effect time! */
  server_contacted = 1;
}

static int do_default (int flags)
{
  unsigned long n;

  command ();
  if (has_body (copy_response ()))
    {
      n = copy_body (c_out, s_in);
      if (log_list_size && (flags & CMD_SIZE))
        syslog (LLEV, "%s: bytes=%lu", cmd, n);
    }
  return 0;
}

/* AUTHINFO USER - called from do_auth. */

static int do_auth_user (int flags)
{
  static char server_buf[512];
  static char user_buf[512];
  int code;
  size_t sep, start, i, n;

  /* Check if connected to server already. */
  if (server_contacted)
    {
      do_default (flags);
      return 0;
    }

  /* Move onto the start of the user name. */
  i = cmd_length;
  while (input[i] != 0 && WHITE (input[i]))
    ++i;

  start = i;
  sep = i - 1;

  while (input[i] != 0 && !WHITE (input[i]))
    {
      if (input[i] == '@')
	sep = i;
      ++i;
    }

  /* Get the server name. */
  n = i - sep - 1;
  if (n >= sizeof (server_buf))
    {
      respond ("501 server name too long");
      return 0;
    }
  memcpy (server_buf, input + sep + 1, n);
  server_buf[n] = 0;
  server = server_buf;

  /* Get the user name. */
  if (sep > start)
    {
      n = sep - start;
      if (n >= sizeof (user_buf))
	{
	  respond ("501 user name too long");
	  return 0;
	}
      memcpy (user_buf, input + start, n);
      user_buf[n] = 0;
      user = user_buf;
    }

  /* Check allowed to connect to server. */
  if (!config_destinations (server, confp))
    {
      /* Not allowed to connect to server. */
      /* Ask for more auth - will fail on USER (as no server length). */
      respond ("381 More authentication information required");
      return 0;
    }

  /* Do the initial connection to the server. */
  open_server ();

  if (user)
    {
      /* Auth - discard the banner from the far end. */
      discard_banner ();

      /* Want to ignore everything from the @ sign. */
      input[sep] = 0;
      input_length = sep;

      /* Now pass the AUTH command through. */
      command ();

      /* And get the response */
      code = copy_response ();

      if (log_auth)
	syslog (LLEV, "AUTHINFO USER: code=%d", code);
    }
  else
    {
      /* No auth - pass the returned banner to the user. */
      display_banner ();
    }
  return 0;
}

/* AUTHINFO PASS - called from do_auth. */

static int do_auth_pass (int flags)
{
  int code;

  if (!server_contacted)	/* First check we're connected to the server */
    respond ("482 Authentication rejected");
  else if (user == NULL)	/* Check if need to pass this through */
    respond ("281 Authentication accepted");
  else
    {
      /* Pass the auth command through to the server. */
      command ();
      code = copy_response ();
      if (log_auth)
        syslog (LLEV, "AUTHINFO PASS: code=%d", code);
    }
  return 0;
}

/* Hash table of AUTH subcommands.  This include file is generated
   from auth.tab by maketable. */

#include "authinfo.h"

/* AUTHINFO */

static int do_auth (int flags)
{
  const char *s = input + cmd_length;
  const char *e;
  size_t len;
  const struct hash_entry *he;
  int (*handler)(int);

  while (*s != 0 && WHITE (*s))
    ++s;
  e = s;
  while (*e != 0 && !WHITE (*e))
    ++e;
  len = e - s;

  if (len == 0 || len >= sizeof (cmd))
    respond ("482 Authentication rejected");
  else
    {
      upper_copy (cmd, s, len);
      cmd[len] = 0;
      cmd_length = e - input;
      he = find_hash_entry2 (&authinfo_descr, cmd, len);
      if (he == NULL)
        respond ("482 Authentication rejected");
      else
        {
          handler = he->handler;
          ALWAYS_ASSERT (AUTHINFO_HANDLER (handler));
          handler (he->flags);
	}
    }
  return 0;
}

static void do_null (void)
{
  syslog (LLEV, "empty command");
}

static void do_unknown (void)
{
  syslog (LLEV, "unknown command: %.512s", input);
  respond ("500 command not recognized");
}

/* ARTICLE, BODY, and HEAD. */

static int do_article (int flags)
{
  command ();
  if (has_body (copy_response ()))
    {
      group_bytes += copy_body (c_out, s_in);
      ++group_articles;
    }
  return 0;
}

/* GROUP */

static int do_group (int flags)
{
  const char *s = input + cmd_length;
  const char *e, *msg;
  size_t len;
  int delay;

  while (*s != 0 && WHITE (*s))
    ++s;
  e = s;
  while (*e != 0 && !WHITE (*e))
    ++e;
  len = e - s;
  if (*e != 0 || len >= sizeof (group_name))
    respond ("501 command syntax error");
  else
    {
      log_group ();
      memcpy (group_name, s, len);
      group_name[len] = 0;
      group_bytes = 0; group_articles = 0;
      if (check_group (group_name, &msg, &delay) != 0)
	{
	  respond (msg);
	  if (delay > 0)
	    sleep (delay);
	  else if (delay < 0)
	    quit (1);
	}
      else
        do_default (flags);
    }
  return 0;
}

/* Hash table of NNTP extensions (for LIST EXTENSIONS).  This include
   file is generated from ext.tab by maketable. */

#include "ext.h"

/* LIST EXTENSIONS */

static int do_list_extensions (int flags)
{
  const struct hash_entry *he;
  const char *s;
  size_t len;
  int code;

  command ();
  code = copy_response ();
  if (code == 202 || code == 215)
    {
      for (;;)
	{
	  input_length = read_line (s_in, input, sizeof (input));
	  if (strcmp (input, ".") == 0)
	    break;
	  s = input; len = input_length;
	  if (*s == ' ')
	    {
	      ++s; --len;
	    }
	  if (len >= sizeof (cmd))
	    he = NULL;
	  else
	    {
	      upper_copy (cmd, s, len);
	      cmd[len] = 0;
	      he = find_hash_entry2 (&ext_descr, cmd, len);
	    }
	  if (he == NULL)
	    syslog (LLEV, "unknown extension: %.512s", input);
	  else if (he->flags != 0)
	    {
	      emo_write (c_out, input, input_length);
	      emo_write (c_out, "\r\n", 2);
	    }
	}
      respond (".");
    }
  return 0;
}

/* Hash table of LIST subcommands.  This include file is generated
   from list.tab by maketable. */

#include "list.h"

/* LIST */

static int do_list (int flags)
{
  const char *s = input + cmd_length;
  const char *e;
  size_t len;
  const struct hash_entry *he;
  int (*handler)(int);

  while (*s != 0 && WHITE (*s))
    ++s;
  e = s;
  while (*e != 0 && !WHITE (*e))
    ++e;
  len = e - s;
  if (len == 0)
    do_default (flags);		/* LIST ACTIVE */
  else
    {
      if (len >= sizeof (cmd))
	do_unknown ();
      else
	{
	  upper_copy (cmd, s, len);
	  cmd[len] = 0; cmd_length = e - input;
	  he = find_hash_entry2 (&list_descr, cmd, len);
	  if (he == NULL)
	    do_unknown ();
	  else
	    {
	      handler = he->handler;
	      ALWAYS_ASSERT (LIST_HANDLER (handler));
	      handler (he->flags);
	    }
	}
    }
  return 0;
}

/* LISTGROUP */

static int do_listgroup (int flags)
{
  int code;
  unsigned long n;

  command ();
  code = copy_response ();
  if (code == 211 || has_body (code))
    {
      n = copy_body (c_out, s_in);
      if (log_list_size)
	syslog (LLEV, "LISTGROUP: bytes=%lu", n);
    }
  return 0;
}

/* POST and IHAVE */

static int do_post (int flags)
{
  unsigned long n;
  int code;

  command ();
  code = copy_response ();
  if (code == 335 || code == 340)
    {
      n = copy_body (s_out, c_in);
      code = copy_response ();
      if (log_post)
	{
	  if (code == 235 || code == 240)
	    syslog (LLEV, "%s: bytes=%lu", cmd, n);
	  else
	    syslog (LLEV, "%s: error=%d bytes=%lu", cmd, code, n);
	}
    }
  else if (log_post)
    syslog (LLEV, "%s: error=%d", cmd, code);
  return 0;
}

/* QUIT */

static int do_quit (int flags)
{
  if (server_contacted)
    do_default (flags);
  quit (0);
  return 0;
}

/* Specific configuration for nntp-gw. */

static void config_nntp_gw (Cfg *confp)
{
  Cfg *cf;
  int i;

  if ((cf = cfg_get ("server", confp)) != NULL)
    {
      if (cf->argc != 1 && cf->argc != 2)
	{
	  syslog (LLEV,
		  "fwtkcfgerr: server must have one or two parameters, "
		  "line %d", cf->ln);
	  exit (1);
	}
      server = cf->argv[0];
      if (cf->argc == 2)
	server_port = str_to_port (cf->argv[1]);
    }

  if ((cf = cfg_get ("log", confp)) != NULL)
    for (i = 0; i < cf->argc; ++i)
      {
	if (strcmp (cf->argv[i], "command") == 0)
	  log_command = 1;
	else if (strcmp (cf->argv[i], "response") == 0)
	  log_response = 1;
	else if (strcmp (cf->argv[i], "list-size") == 0)
	  log_list_size = 1;
	else if (strcmp (cf->argv[i], "article-summary") == 0)
	  log_article_summary = 1;
	else if (strcmp (cf->argv[i], "post") == 0)
	  log_post = 1;
        else if (strcmp (cf->argv[i], "auth") == 0)
          log_auth = 1;
	else
	  {
	    syslog (LLEV, "fwtkcfgerr: invalid parameter for log, line %d",
		    cf->ln);
	    exit (1);
	  }
      }
}

/* Catch SIGPIPE.  SIGPIPE is generated when writing to a closed pipe
   or socket. */

static void catch_sigpipe (int signo)
{
  syslog (LLEV, "attempt to write to closed connection");
  quit (1);
}

/* Hash table of NNTP commands.  This include file is generated from
   command.tab by maketable. */

#include "command.h"

/* Read commands from the client and pass them to the server. */

static void cmd_loop (void)
{
  size_t len;
  const struct hash_entry *he;
  int (*handler)(int);

  for (;;)
    {
      input_length = read_line (c_in, input, sizeof (input));
      len = 0;
      while (input[len] != 0 && !WHITE (input[len]))
	++len;
      if (len >= sizeof (cmd))
	do_unknown ();
      else if (len == 0)
        do_null ();
      else
	{
	  upper_copy (cmd, input, len);
	  cmd[len] = 0; cmd_length = len;
	  he = find_hash_entry2 (&command_descr, cmd, len);
	  if (he == NULL)
	    do_unknown ();
	  else if (!server_contacted && !(he->flags & CMD_NOAUTH))
	    respond ("450 Authorization required for this command");
	  else
	    {
	      handler = he->handler;
	      ALWAYS_ASSERT (COMMAND_HANDLER (handler));
	      handler (he->flags);
	    }
	}
    }
}

/* This is where nntp-gw starts. */

int main (int argc, char *argv[])
{
  char *config_name = "nntp-gw";

  if (argc >= 2 && strcmp (argv[1], "-check") == 0)
    {
      check_snprintf (0);
      exit (0);
    }

  openlog ("nntp-gw", LOG_PID|LOG_NDELAY, LFAC);

  if (argc >= 3 && strcmp (argv[1], "-daemon") == 0)
    {
      int port = str_to_port (argv[2]);
      argc -= 2; argv += 2;
      syslog (LLEV, "Starting daemon mode on port %d", port);
      do_daemon (port);
    }

  if (argc >= 2 && argv[1][0] != '-')
    {
      config_name = argv[1];
      argc -= 1; argv += 1;
    }

  if (argc != 1)
    {
      syslog (LLEV, "Bad command line");
      exit (1);
    }

  if ((confp = cfg_read (config_name)) == (Cfg *)-1)
    exit (1);
  config_groupid (confp, 1);
  config_userid (confp, 1);
  config_nntp_gw (confp);

  if (peername (0, rladdr, riaddr, sizeof (riaddr)) != 0)
    {
      syslog (LLEV, "fwtksyserr: cannot get peer name: %m");
      exit (1);
    }

  if (config_hosts (confp, rladdr, riaddr) == NULL)
    exit (1);

  c_in = emi_fdopen (0, 4096);
  c_out = emo_fdopen (0, 4096);
  if (c_in == NULL || c_out == NULL)
    {
      syslog (LLEV, "fdopen() failed: %m");
      exit (1);
    }
  emi_timeout (c_in, PROXY_TIMEOUT);
  emo_timeout (c_out, PROXY_TIMEOUT);

  /* Install signal handler. */

  signal (SIGPIPE, catch_sigpipe);

  if (server)
    {
      /* Config has "server" parameter - so we want to connect now. */
      open_server ();
      display_banner ();
    }
  else
    {
      /* Will connect when the user sends AUTHINFO. */
      respond ("200 Hello you can post or not");
    }

  /* Now sit there processing commands. */
  cmd_loop ();
  return 0;
}
