/* pop-gw.c */

/* Copyright 1997-2000 by Eberhard Mattes <em-gw@windhager.de>
   Donated to the public domain.  No warranty.

   Version 0.1	1997-04-09 Initial version
   Version 0.2	1997-08-22 -user option for permit-hosts
   Version 0.3	1997-08-23 APOP support added
   Version 0.4	1998-09-03 Support multiple "user" lines; check_snprintf()
   Version 0.5	1999-03-17 Use copy_bidir()
   Version 0.6	2000-04-17 Use namatch(); "user-limit", "pass-limit"
   Version 0.7	2000-05-31 copy_bidir() has an additional argument
   Version 0.8	2000-06-22 -apop option for permit-hosts and user */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#include "firewall.h"
#include "libemfw.h"
#include "libemtn.h"

#define SERVER_PORT	110

static Cfg *confp;
static Cfg *cf_hosts;
static const char *server;
static char timestamp[128];
static char user[256+1];
static char pass[256+1];
static char riaddr[512];
static char rladdr[512];
static char input[512];
static tnconn *tc, *ts;
static int server_fd;
static int user_limit = 31;
static int pass_limit = 31;

/* firewall.h lacks some declarations. */

int conn_server (const char *, int, int, char *);
int str_to_port (char *);
int do_daemon (int);
int peername (int, char *, char *, int);
int namatch (const char *, const char *);

/* Send positive answer to client. */

static void ok (void)
{
  tn_puts (tc, "+OK\r\n");
}

/* Send negative answer to client: authorization failed. */

static void auth_failed (void)
{
  tn_puts (tc, "-ERR authorization failed\r\n");
  exit (1);
}

/* Send command to server and check for positive answer. */

static void command (const char *cmd, const char *arg)
{
  tn_printf (ts, "%s %s\r\n", cmd, arg);
  tn_gets (ts, input, sizeof (input), PROXY_TIMEOUT, 0);
  if (input[0] != '+')
    auth_failed ();
}

/* Result of check_user(). */

enum cu_result
{
  cu_ok, cu_apop_required, cu_fail
};

/* Check user name against configuration file. */

static enum cu_result check_user (void)
{
  Cfg *cf;
  int i, apop, match;

  /* Try the "-user" option of the "hosts" attribute. */

  cf = cf_hosts;
  i = 0; apop = 0; match = 0;
  while (i < cf->argc)
    {
      const char *s = cf->argv[i++];
      if (strcmp (s, "-apop") == 0)
	apop = 1;
      else if (strcmp (s, "-user") == 0)
	{
	  while (i < cf->argc && cf->argv[i][0] != '-')
	    {
	      if (!match && namatch (cf->argv[i], user))
		match = 1;
	      ++i;
	    }
	}
      else if (*s == '-')
	{
	  syslog (LLEV, "fwtkcfgerr: unknown option in line %d", cf->ln);
	  exit (1);
	}
    }

  if (match)
    return apop ? cu_apop_required : cu_ok;

  /* Try the "user" attribute. */

  cf = cfg_get ("user", confp);
  while (cf != (Cfg *)0)
    {
      i = 0;
      while (i < cf->argc)
	{
	  const char *s = cf->argv[i++];
	  if (strcmp (s, "-apop") == 0)
	    apop = 1;
	  else if (*s == '-')
	    {
	      syslog (LLEV, "fwtkcfgerr: unknown option in line %d", cf->ln);
	      exit (1);
	    }
	  else
	    {
	      if (!match && namatch (s, user))
		match = 1;	/* Keep looking for -apop */
	      ++i;
	    }
	}
      if (match)
	return apop ? cu_apop_required : cu_ok;
      cf = cfg_get ("user", (Cfg *)0);
    }
  return cu_fail;
}

/* Implement transaction state of the POP3 protocol.  This function
   never returns. */

static void transaction_state (void)
{
  int r;

  ok ();
  r = copy_bidir (0, server_fd, PROXY_TIMEOUT, CB_EOF1 | CB_EOF2);
  if (r > 0)
    syslog (LLEV, "timeout");
  exit (r != 0 ? 1 : 0);
}

/* Perform the "QUIT" command. */

static void do_quit (void)
{
  syslog (LLEV, "QUIT");
  ok ();
  exit (0);
}

/* Perform the "USER" command. */

static void do_user (const char *s)
{
  enum cu_result r;

  while (*s == ' ') ++s;
  syslog (LLEV, "USER %s", s);
  if (*s == 0 || strlen (s) > (size_t)user_limit)
    {
      tn_puts (tc, "-ERR syntax error\r\n");
      return;
    }
  strcpy (user, s);
  r = check_user ();
  if (r == cu_apop_required)
    auth_failed ();		/* Attempt to prevent sending PASS */
  ok ();			/* Even for cu_fail! */
}

/* Perform the "PASS" command.  This function never returns. */

static void do_pass (const char *s)
{
  enum cu_result r;

  syslog (LLEV, "PASS");
  if (user[0] == 0)
    {
      tn_puts (tc, "-ERR need USER\r\n");
      return;
    }
  while (*s == ' ') ++s;
  if (*s == 0 || strlen (s) > (size_t)pass_limit)
    {
      tn_puts (tc, "-ERR syntax error\r\n");
      return;
    }
  strcpy (pass, s);
  r = check_user ();
  if (r != cu_ok)
    auth_failed ();
  command ("USER", user);
  command ("PASS", pass);
  transaction_state ();
}

/* Perform the "APOP" command.  This function never returns. */

static void do_apop (const char *s)
{
  int i;
  const char *h;
  enum cu_result r;

  while (*s == ' ') ++s;
  syslog (LLEV, "APOP %s", s);
  i = 0;
  while (s[i] != 0 && s[i] != ' ')
    ++i;
  h = s + i;
  while (*h == ' ') ++h;
  if (i < 1 || i > user_limit || strlen (h) != 32 || timestamp[0] == 0)
    {
      tn_puts (tc, "-ERR syntax error\r\n");
      return;
    }
  memcpy (user, s, i);
  user[i] = 0;
  r = check_user ();
  if (r != cu_ok && r != cu_apop_required)
    auth_failed ();
  command ("APOP", s);
  transaction_state ();
}

/* Open a connection to the server.  We do this before sending the
   banner to the client as we need the timestamp for APOP. */

static void open_server (void)
{
  char *left, *right;

  if ((server_fd = conn_server (server, SERVER_PORT, 0, (char *)0)) < 0)
    {
      syslog (LLEV, "cannot connect to server %.512s/%d: %m",
	      server, (int)SERVER_PORT);
      exit (1);
    }
  ts = tn_init (server_fd, 0);
  if (ts == NULL)
    exit (1);

  tn_gets (ts, input, sizeof (input), PROXY_TIMEOUT, 0);
  if (input[0] != '+')
    {
      syslog (LLEV, "server failed: %.512s", input);
      exit (1);
    }
  left = strchr (input, '<');
  if (left != NULL)
    {
      right = strchr (left + 1, '>');
      if (right != NULL)
	{
	  size_t len = right + 1 - left;
	  if (len >= sizeof (timestamp))
	    {
	      syslog (LLEV, "timestamp too long");
	      exit (1);
	    }
	  memcpy (timestamp, left, len);
	  timestamp[len] = 0;
	}
    }
  if (timestamp[0] != 0)
    tn_printf (tc, "+OK %s\r\n", timestamp);
  else
    ok ();
}

/* Specific configuration for pop-gw. */

static void config_pop_gw (Cfg *confp)
{
  Cfg *cf;

  if ((cf = cfg_get ("server", confp)) == (Cfg *)0)
    {
      syslog (LLEV, "fwtkcfgerr: `server' required");
      exit (1);
    }
  config_arg_count (cf, 1);
  server = cf->argv[0];

  user_limit = config_int (confp, "user-limit", 1, sizeof (user) - 1,
			   user_limit);
  pass_limit = config_int (confp, "pass-limit", 1, sizeof (pass) - 1,
			   pass_limit);
}

/* This is where pop-gw starts. */

int main (int argc, char *argv[])
{
  if (argc >= 2 && strcmp (argv[1], "-check") == 0)
    {
      check_snprintf (0);
      exit (0);
    }

  openlog ("pop-gw", LOG_PID|LOG_NDELAY, LFAC);

  if (argc >= 3 && strcmp (argv[1], "-daemon") == 0)
    {
      int port = str_to_port (argv[2]);
      argc -= 2; argv += 2;
      syslog (LLEV, "Starting daemon mode on port %d", port);
      do_daemon (port);
    }

  if ((confp = cfg_read ("pop-gw")) == (Cfg *)-1)
    exit (1);
  config_groupid (confp, 1);
  config_userid (confp, 1);
  config_pop_gw (confp);

  if (peername (0, rladdr, riaddr, sizeof (riaddr)) != 0)
    {
      syslog (LLEV, "fwtksyserr: cannot get peer name: %m");
      exit (1);
    }

  cf_hosts = config_hosts (confp, rladdr, riaddr);
  if (cf_hosts == (Cfg *)0)
    exit (1);

  tc = tn_init (0, 0);
  if (tc == NULL)
    exit (1);

  open_server ();

  for (;;)
    {
      tn_gets (tc, input, sizeof (input), PROXY_TIMEOUT, 0);
      if (strncasecmp (input, "quit", 4) == 0 && input[4] == 0)
        do_quit ();
      else if (strncasecmp (input, "user", 4) == 0 && input[4] == ' ')
        do_user (input+4);
      else if (strncasecmp (input, "pass", 4) == 0 && input[4] == ' ')
        do_pass (input+4);
      else if (strncasecmp (input, "apop", 4) == 0 && input[4] == ' ')
        do_apop (input+4);
      else
        tn_puts (tc, "-ERR authorization first\r\n");
    }
}
